@@ -7,7 +7,7 @@ Database schema version control library for SQLAlchemy.
 
 __author__ = 'Chris Miles'
 __copyright__ = '(c) Chris Miles 2008. All rights reserved.'
-__id__ = '$Id: __init__.py 3 2009-02-06 04:05:07Z chris $'
+__id__ = '$Id: __init__.py 6 2009-02-23 08:47:34Z chris $'
 __url__ = '$URL: https://psychofx.com/schemabot/svn/SchemaBot/trunk/schemabot/__init__.py $'
 
 
@@ -31,6 +31,8 @@ default_logger = logging.getLogger("schemabot")
 class SchemaBotError(Exception):
     pass
 
+class ChangeAlreadyDefined(SchemaBotError):
+    pass
 
 
 # ---- Classes ----
@@ -61,6 +63,9 @@ class SchemaChange(object):
         if engine_name is None:
             engine_name = '_default_'
         
+        if engine_name in self.engines:
+            raise ChangeAlreadyDefined("Changes are already defined for engine", engine_name)
+        
         self.engines[engine_name] = dict(
             upgrade = upgrade,
             downgrade = downgrade,
@@ -117,7 +122,7 @@ class SchemaBot(object):
         
         Args:
         
-        `metadata` : an initialise sqlalchemy.MetaData object that is bound to an
+        `metadata` : an initialised sqlalchemy.MetaData object that is bound to an
         sqlalchemy engine.
         
         `create_table` : bool indicating whether to automatically create the
@@ -152,9 +157,14 @@ class SchemaBot(object):
     
     def define_schema(self, version, upgrade=None, downgrade=None, engine_name=None):
         version = int(version)
+        if version == 0:
+            raise SchemaBotError("Schema version 0 is the initial schema state and cannot contain any schema changes.")
         
         changeobj = self.changes.setdefault(version, SchemaChange(log=self.log))
-        changeobj.define_change(engine_name, upgrade, downgrade)
+        try:
+            changeobj.define_change(engine_name, upgrade, downgrade)
+        except ChangeAlreadyDefined, err:
+            raise ChangeAlreadyDefined("Schema version %d has already been defined for engine '%s'" %(version, err[1]))
     
     def get_current_version(self):
         """Fetch the current schema version from the DB (schemabot_version.current_version).
@@ -185,12 +195,12 @@ class SchemaBot(object):
                 raise SchemaBotError("No schema versions are defined.")
                 
             if lowest:
-                version = min(versions)     # smallest version number
+                version = min(versions)     # smallest version number (excluding 0, the initial version)
             else:
                 version = max(versions)     # largest version number
         else:
             version = int(version)          # ensure int
-            if not self.changes.get(version):
+            if version != 0 and not self.changes.get(version):
                 raise SchemaBotError("No schema version defined for version %d." %version)
         
         return version
@@ -200,12 +210,6 @@ class SchemaBot(object):
         with the version specified (or the maximum defined version if version is None).
         
         Returns a 2-tuple (model_version, current_DB_version)
-        
-        # Returns an int between -1 and 1:
-        # 
-        # -1 : DB version < version => schema upgrade required
-        #  0 : DB version == version => no action required
-        #  1 : DB version > version => downgrade required
         """
         model_version = self.model_version(version)
         current_version = self.get_current_version()
@@ -242,13 +246,16 @@ class SchemaBot(object):
         minimum (downgrade) defined versions; or to the explicit version if
         the `version` argument is specified.
         
+        Specify version 0 to downgrade all schema changes.  Version 0 is always
+        the initial schema state.
+        
         All changes are performed within a transaction, so if any fails the
         whole lot will be rolled back.  i.e. The schema change will be all or
         nothing.  With the exception of the note below.
         
-        NOTE: some databases do not support CREATE/DROP/etc within a transaction,
-        so those operations cannot be rolled back.  In those cases, the database
-        may be left in an inconsistent state if an error occurs during upgrade
+        NOTE: some databases (and/or adaptors) do not support CREATE/DROP/etc within
+        a transaction, so those operations cannot be rolled back.  In those cases, the
+        database may be left in an inconsistent state if an error occurs during upgrade
         or downgrade.
         Ref http://groups.google.com/group/sqlalchemy/msg/7756ba6582336f54
         """
